# Copyright (c) 2010-2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import print_function
import re

from . import _compat

# Break a cyclical loop:
def _has_toolkit(toolkit_name):
    from . import has_toolkit
    globals()["_has_toolkit"] = has_toolkit
    return has_toolkit(toolkit_name)
    
##

# This isn't yet a public class.
# It should also derive from ChemFPError
# However, there's an import loop to resolve.
class FingerprintTypeValueError(ValueError):
    pass

######## Parse a type string

# Type strings are of the form
#
#    <name>  (<WS> <argument-name> '=' <argument-value>)*
#
#  where <name> is of the form:
#
#    <base name> ('/' <version>)?
#
# The base name name and version match [!-.0-~]+  (printable ASCII except for "/")
# The keys match [a-zA-Z_][a-zA-Z0-9_]* . The values match [!-~]+

_name_pat = re.compile(r"([!-.0-~]+)(/([!-.0-~]+))?$")

_not_allowed_char_pat = re.compile(r"[^!-.0-~]")
def _contains_special_characters(name):
    return _not_allowed_char_pat.search(name) is not None


def _verify_name(name):
    "Verify that the type string name is correct"
    # Quick check to see if it's in the right form.
    m = _name_pat.match(name)
    if m is not None:
        return m.group(1), m.group(3)

    # Something went wrong. Let's see if I can give a better report.
    if not name:
        raise FingerprintTypeValueError("Unable to use an empty fingerprint type name")
    if not name.strip():
        raise FingerprintTypeValueError("Unable to use a blank fingerprint type name: %r" % (name,))

    n = name.count("/")
    if n == 0:
        if _contains_special_characters(name):
            raise FingerprintTypeValueError("Fingerprint type name contains an unexpected character: %r" % (name,))
        return name, None
    if n == 1:
        base_name, version = name.split("/")
        
        if not base_name:
            raise FingerprintTypeValueError("Missing fingerprint type name before the first '/': %r" % (name,))
        if _contains_special_characters(base_name):
            raise FingerprintTypeValueError("Fingerprint type name contains an unexpected character: %r" % (name,))
        
        if not version:
            raise FingerprintTypeValueError("Missing version after the first '/' of the fingerprint type: %r" % (name,))
        if _contains_special_characters(version):
            raise FingerprintTypeValueError("Fingerprint type version contains an unexpected character: %r" % (name,))

        return base_name, version

    raise FingerprintTypeValueError("Too many '/'s in the fingerprint type name: %r" % (name,))


    
class TokenizedType(object):
    """A tokenized type string

    `name` is the type name, which may or may not have a version
    string. `base_name` is the type name without the version,
    and `version` is the version component, or None if there is
    no version.

    `settings` is a list of argument pairs as the 2-element
    tuple (argument name, argument value), where both the name
    and value are strings.
    """
    
    def __init__(self, base_name, version, settings):
        self.base_name = base_name
        self.version = version
        if version is None:
            self.name = base_name
        else:
            self.name = base_name + "/" + version
        self.settings = settings

def tokenize_type_string(type):
    """Convert a type string into a TokenizedType instance

    This is most easily understood as an example.
    
    >>> tokens = tokenize_type_string("ABC/123 a=1 b=xyz")
    >>> tokens.name, tokens.base_name, tokens.version
    ('ABC/123', 'ABC', '123')
    >>> tokens.settings
    [('a', '1'), ('b', 'xyz')]
    """
    base_name, version, settings = _split_type_string(type)
    
    # It starts off correct. Now parse the name/value settings.

    seen = set()
    tokens = []
    
    for term in settings:
        try:
            left, right = term.split("=")
        except ValueError:
            raise FingerprintTypeValueError("Term %r in type %r must have one and only one '='" %
                                                (term, type))
        if left in seen:
            raise FingerprintTypeValueError("Duplicate parameter %r in type %r" % (left, type))

        seen.add(left)
        tokens.append( (left, right) )
    
    return TokenizedType(base_name, version, tokens)

def _split_type_string(type):
    """Break up a type string into its (base name, version, settings)"""
    if not type:
        if type is None:
            raise FingerprintTypeValueError("Must specify a fingerprint type string")
        raise FingerprintTypeValueError("Unable to use an empty fingerprint type")
    if type.isspace():
        raise FingerprintTypeValueError("Empty fingerprint type (%r)" % (type,))
    if type[:1].isspace():
        raise FingerprintTypeValueError("Fingerprint type must not begin with whitespace: %r" % (type,))

    terms = type.split()
    if not terms:
        raise AssertionError("Unexpected empty fingerprint type (%r)" % (type,))

    name = terms[0]
    base_name, version = _verify_name(name)
    return base_name, version, terms[1:]


######## Manage information about how to convert string parameters into actual arguments

# A FingerprintArguments instance contains information about the
# arguments for a type, including a help message and the ways to
# convert argument values to and from a string, plus some extra
# information for argparse.ArgumentParser.add_argument().

class FingerprintArgument(object):
    """Information about a fingerprint type argument and converters to/from a string

    :param name: the argument name
    :type name: string
    :param default: default value for the argument
    :type default: any
    :param help: help message describing the name (used for command-line arguments)
    :type help: string
    :param decoder: function to convert a string to its actual value
    :type decoder: callable(str) which raises ValueError on error
    :param encoder: function to convert a value to a string
    :type encoder: callable(value) which raises ValueError on error
    :param action: value used for argparse's add_argument
    :type action: string or None 
    :param metavar: value used for argparse's add_argument
    :type metavar: string or None
    """
    def __init__(self, name, default, help, decoder, encoder=str,
                 action=None, metavar=None):
        self.name = name
        self.default = default
        self.help = help
        self.decoder = decoder
        self.encoder = encoder
        self.action = action
        self.metavar = metavar


class FingerprintArguments(object):
    """Information about how to convert fingerprint type arguments

    `arguments` is a dictionary mapping an argument name to its
    FingerprintArgument.

    """
    def __init__(self, arguments={}):
        self.arguments = arguments.copy()

    # TODO: add more of the dictionary API?
    def __getitem__(self, name):
        return self.arguments[name]

    def __contains__(self, name):
        return name in self.arguments
        

    def select_names(self, names):
        """Make a new FingerprintArguments() using the selected argument names

        `names` is either list of argument names or a string containing
        whitespace separated names.

        :param name: the argument names for the new FingerprintArguments
        :type name: list of strings, or a whitespace-separated string with names
        :returns: a new FingerprintArguments using only the selected names
        """
        if isinstance(names, _compat.filename_types):
            names = names.split()
        new_arguments = {}
        for name in names:
            new_arguments[name] = self.arguments[name]
        return FingerprintArguments(new_arguments)

    def add_argument(self, name, default, help, decoder, 
                     encoder=str, action=None, metavar=None,
                     add_default_to_help=True):
        """Add a new argument

        Add a new argument named `name` with default value `default`.
        If `add_default_to_help` is True then the `help` message is
        combined with information about the default value to make
        the complete help message. Otherwise the help string is used
        as-is.
        
        Argument values are externally represented as strings.
        The `decoder` and `encoder` callables convert the string
        to/from the actual value. The `decoder` may raise a
        ValueError if there's an error. (For example, Python's
        `float` builtin is a valid decoder.)

        The `action` and `metavar` values are passed unchanged
        to argparse.
        
        :param name: the name of the argument
        :type name: string
        :param decoder: a callable which converts a string to a value
                   or raises a ValueError if there's an error
        :param default: the default value for the argument
        :param help: text passed to argparse
        :param encoder: a callable which converts a valid value to
                   a string
        :param action: action value passed to argparse
        :param metavar: metavar value passed to argparse
        :param add_default_to_help: if True, combine help and encoder(default)
                   to make the actual help message
        """
        if name in self.arguments:
            raise FingerprintTypeValueError("Argument %r already specified" % (name,))

        if add_default_to_help:
            help = "%s (default=%s)" % (help, encoder(default))
            
        self.arguments[name] = FingerprintArgument(name, default, help,
                                                   decoder, encoder, action, metavar)


    def decode_type_parameters(self, parameters):
        """Convert name/encoded_value pairs into a fingerprint_args dict

        `parameters` is an iterable of (name, string encoded value) pairs.
        Unknown names and duplicate names raise a ValueError. Default
        values are used for missing arguments.

        :param parameters: an iterable of (name, string encoded value) pairs
        :returns: a dictionary of names and decoded values
        """
        arguments = self.arguments

        # Set up the defaults
        kwargs = dict( (name, arg.default) for (name, arg) in arguments.items() )

        # Go through the parameters
        seen = set()
        for name, value in parameters:
            # Check for unknown and duplicate names
            if name not in arguments:
                raise FingerprintTypeValueError("Unsupported fingerprint parameter name %r" % (name,))
            if name in seen:
                raise FingerprintTypeValueError("Duplicate fingerprint parameter name %r" % (name,))
            seen.add(name)
            arg = arguments[name]
            try:
                kwargs[name] = arg.decoder(value)
            except ValueError as err:
                raise FingerprintTypeValueError("Unable to parse fingerprint parameter %r value %r: %s" % (name, value, err))

        return kwargs

    def merge_with_defaults(self, fingerprint_args):
        """Fill in missing parameters in a dictionary of fingerprint arguments.

        `fingerprint_args` is a dictionary containing parameter names and
        (decoded) values. The output is a new dictionary containing the names
        and values of the input dictionary, plus default values for the missing
        parameter names.

        Raises a ValueError if one of the `fingerprint_args` dictionary keys
        is not a known parameter.
        """
        arguments = self.arguments
        
        # Set up the defaults
        kwargs = dict( (name, arg.default) for (name, arg) in arguments.items())

        for name, value in fingerprint_args.items():
            if name not in arguments:
                raise FingerprintTypeValueError("Unsupported fingerprint parameter %r (=%r)" % (name, value))
            # Allow people to specify 'None' to mean "use the default value"
            if value is not None:
                kwargs[name] = value
        
        return kwargs

    def encode_fingerprint_args(self, fingerprint_args):
        """Convert parameter names and values into a dict of name and string-encoded values

        `fingerprint_args` is a dictionary of parameter names and (decoded) values.
        The result is a new dictionary with the same names and string encoded values.

        Raises a ValueError if one of the `fingerprint_args` dictionary keys
        is not a known parameter, or if one of the values could not be encoded.
        """
        arguments = self.arguments
        d = {}
        for (name, value) in fingerprint_args.items():
            try:
                encoder = arguments[name].encoder
            except KeyError:
                raise FingerprintTypeValueError("Unsupported fingerprint argument %r (=%r)" % (name, value))
            try:
                d[name] = encoder(value)
            except ValueError as err:
                raise FingerprintTypeValueError(str(err) + (" in argument %s" % (name,)))
        return d

# Helper functions to validate fingerprint type arguments

def positive_int(s):
    # Don't do int(s) because that allows "+3" and " 3 ", which I don't want
    if not s.isdigit():
        raise FingerprintTypeValueError("must be 1 or greater")
    i = int(s)
    if i == 0:
        raise FingerprintTypeValueError("must be 1 or greater")
    return i

def nonnegative_int(s):
    if not s.isdigit():
        raise FingerprintTypeValueError("must be 0 or greater")
    return int(s)

def zero_or_one(s):
    if s == "0":
        return 0
    if s == "1":
        return 1
    raise FingerprintTypeValueError("must be 0 or 1")

def encode_bool_as_zero_or_one(val):
    if val:
        return "1"
    return "0"

def power_of_2_at_least_32(s):
    if not s.isdigit():
        raise FingerprintTypeValueError("must a power of 2 and at least 32")
    i = int(s)
    if i < 32 or bin(i).count("1") != 1:
        raise FingerprintTypeValueError("must a power of 2 and at least 32")
    return i
    
def fromAtom_list(s):
    t = s.strip()
    if not t:
        # Should I allow the empty string?
        # If so, does that mean no atoms or all atoms?
        raise FingerprintTypeValueError(
            "must contain a comma-separated list of atom indices")

    terms = t.split(",")
    seen = set()
    for term in terms:
        if not term.isdigit():
            raise FingerprintTypeValueError(
                "term %r must be a non-negative integer"
                % (term,))
        seen.add(int(term))
    return sorted(seen)
    

######## The Fingerprint Family classes

class BaseFingerprintFamily(object):
    name = None
    toolkit = None
    def __call__(self, fingerprint_kwargs=None):
        raise NotImplementedError("must be implemented in the subclass")

    def from_kwargs(self, fingerprint_kwargs=None):
        raise NotImplementedError("must be implemented in the subclass")

    def from_text_settings(self, settings=None):
        raise NotImplementedError("must be implemented in the subclass")

    def get_kwargs_from_text_settings(self, settings=None):
        raise NotImplementedError("must be implemented in the subclass")


# The BaseFingerprintFamily is the "real" base class.
# I don't know why I have a BaseFingerprintFamily. Remove it? XXX

class FingerprintFamily(BaseFingerprintFamily):
    """A FingerprintFamily is used to create a FingerprintType or get information about its parameters

    Two reasons to use a FingerprintFamily (instead of using
    :func:`chemfp.get_fingerprint_type` or :func:`chemfp.get_fingerprint_type_from_text_settings`) are:

    * figure out the default arguments;
    * given a text settings or parameter dictionary, use the keys from the default
      argument keys to remove other parameters before creating a FingerprintType
      (otherwise the creation function will raise an exception)

    All fingerprint families have the following attributes:

    * name - the type name, including version
    * toolkit - the toolkit API for the underlying chemistry toolkit, or None
    
    """
    def __init__(self, fingerprint_class):
        """The constructor is not part of the public API."""
        self._fingerprint_class = fingerprint_class
        self._arguments = getattr(self._fingerprint_class, "_arguments", None)

    def __repr__(self):
        """Return a string like 'FingerprintFamily(<RDKit-Fingerprint/2>)'"""
        return "FingerprintFamily(<%s>)" % (self._fingerprint_class.name,)

    @property
    def name(self):
        """The full fingerprint name, including the version"""
        return self._fingerprint_class.name

    @property
    def base_name(self):
        """The base fingerprint name, without the version"""
        return self.name.split("/")[0]
        
    @property
    def version(self):
        """The fingerprint version"""
        return self.name.split("/")[1]
    
    @property
    def toolkit(self):
        """The toolkit used to implement this fingerprint, or None"""
        return self._fingerprint_class.toolkit

    def __call__(self, **fingerprint_kwargs):
        """Create a fingerprint type; keyword arguments can override the defaults

        The argument values are native Python values, not string-encoded values::

          >>> import chemfp
          >>> family = chemfp.get_fingerprint_family("RDKit-Fingerprint")
          >>> fptype = family()
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=2048 nBitsPerHash=2 useHs=1'
          >>> fptype = family(fpSize=1024)
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=1024 nBitsPerHash=2 useHs=1'

        The function will raise an exception for unknown arguments.

        :param fingerprint_kwargs: the fingerprint parameters
        :returns: an object implementing the :class:`chemfp.types.FingerprintType` API
        """
        return self.from_kwargs(fingerprint_kwargs)

    def from_kwargs(self, fingerprint_kwargs=None):
        """Create a fingerprint type; items in the *fingerprint_kwargs* dictionary can override the defaults

        The dictionary values are native Python values, not string-encoded values::

          >>> import chemfp
          >>> family = chemfp.get_fingerprint_family("RDKit-Fingerprint")
          >>> fptype = family()
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=2048 nBitsPerHash=2 useHs=1'
          >>> fptype = family.from_kwargs({"fpSize": 1024})
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=1024 nBitsPerHash=2 useHs=1'

        The function will raise an exception for unknown arguments.

        :param fingerprint_kwargs: the fingerprint parameters
        :type fingerprint_kwargs: a dictionary where the values are Python objects
        :returns: an object implementing the :class:`chemfp.types.FingerprintType` API
        """
        if fingerprint_kwargs is None:
            fingerprint_kwargs = {}

        if self._arguments is not None:
            fingerprint_kwargs = self._arguments.merge_with_defaults(fingerprint_kwargs)

        return self._fingerprint_class(fingerprint_kwargs)
        ## except TypeError as err:
        ##     raise ValueError("Unable to create a %r fingerprint type: %s"
        ##                      % (self._fingerprint_class.name, err))

    def from_text_settings(self, settings=None):
        """Create a fingerprint type; *settings* is a dictionary with string-encoded value that can override the defaults

        The dictionary values are string-encoded values, not native Python values.
        This function exists to help handle command-line arguments and setting files.::

          >>> import chemfp
          >>> family = chemfp.get_fingerprint_family("RDKit-Fingerprint")
          >>> fptype = family.from_text_settings()
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=2048 nBitsPerHash=2 useHs=1'
          >>> fptype = family.from_text_settings({"fpSize": "1024"})
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=1024 nBitsPerHash=2 useHs=1'

        The function will raise an exception for unknown arguments.

        :param settings: the fingerprint text settings
        :type settings: a dictionary where the values are string-encoded
        :returns: an object implementing the :class:`chemfp.types.FingerprintType` API
        """
        if settings is None:
            return self._fingerprint_class(self._arguments.merge_with_defaults({}))
        fingerprint_kwargs = self.get_kwargs_from_text_settings(settings)
        return self._fingerprint_class(fingerprint_kwargs)

    def get_kwargs_from_text_settings(self, settings=None):
        """Convert a dictionary of string-encoded fingerprint parameters into native Python values

        String-encoded values ("text settings") can come from the command-line,
        a configuration file, a web reqest, or other text sources. The fingerprint
        types need actual Python values. This method converts the first to the second::

          >>> import chemfp
          >>> family = chemfp.get_fingerprint_family("RDKit-Fingerprint")
          >>> family.get_kwargs_from_text_settings()
          {'maxPath': 7, 'fpSize': 2048, 'nBitsPerHash': 2, 'minPath': 1, 'useHs': 1}
          >>> family.get_kwargs_from_text_settings({"fpSize": "128", "maxPath": "5"})
          {'maxPath': 5, 'fpSize': 128, 'nBitsPerHash': 2, 'minPath': 1, 'useHs': 1}

        :param settings: the fingerprint text settings
        :type settings: a dictionary where the values are string-encoded
        :returns: an dictionary of (decoded) fingerprint parameters
        """
        if self._arguments is None:
            if settings:
                raise FingerprintTypeValueError("This fingerprint does not take settings")
            return {}
        if settings is None:
           settings = {}
        return self._arguments.decode_type_parameters(settings.items())

    def get_defaults(self):
        """Return the default parameters as a dictionary

        The dictionary values are native Python objects::

          >>> import chemfp
          >>> family = chemfp.get_fingerprint_family("RDKit-Fingerprint")
          >>> family.get_defaults()
          {'maxPath': 7, 'fpSize': 2048, 'nBitsPerHash': 2, 'minPath': 1, 'useHs': 1}

        :returns: an dictionary of fingerprint parameters
        """
        if self._arguments is None:
            return {}
        return self._arguments.merge_with_defaults({})


########

class NoFingerprintParametersMixin(object):
    def __init__(self, fingerprint_kwargs):
        if fingerprint_kwargs:
            raise FingerprintTypeValueError("This fingerprint does not support parameters")
        super(NoFingerprintParametersMixin, self).__init__(fingerprint_kwargs)

class BaseFingerprintType(object):
    name = None
    toolkit = None
    software = None

    def __init__(self, fingerprint_kwargs):
        self.fingerprint_kwargs = fingerprint_kwargs
        base_name, mid, version = self.name.partition("/")
        self.base_name = base_name
        if not mid:
            version = None
        self.version = version
    
    def get_type(self):
        """Get the full type string (name and parameters) for this fingerprint type

        :returns: a canonical fingerprint type string, including its parameters
        """
        raise NotImplementedError("Must be implemented in a derived class")

    def make_fingerprinter(self):
        """Make a 'fingerprinter'; a callable which takes a molecule and returns a fingerprint

        :returns: a function object which takes a molecule and return a fingerprint
        """
        raise NotImplementedError("Must be implemented in the derived class")


    def read_molecule_fingerprints(self, source, format=None, id_tag=None,
                                    reader_args=None, errors="strict", location=None):
        """Read fingerprints from a structure source as a FingerprintIterator

        Iterate through the *format* structure records in *source*. If *format*
        is None then auto-detect the format based on the *source*. Use the
        fingerprint type to compute the fingerprint. For SD files, use *id_tag*
        to get the record id from the given SD tag instead of the title line.

        The *reader_args* dictionary parameters depend on the toolkit and format.
        For details see the docstring for ``self.toolkit.read_molecules``.

        The *errors* parameter specifies how to handle errors. "strict" raises
        an exception, "report" sends a message to stderr and goes to the next
        record, and "ignore" goes to the next record.

        The *location* parameter takes a Location instance. If None then a default
        Location will be created.
        

        :param source: the structure source
        :type source: a filename, file object, or None to read from stdin
        :param format: the input structure format
        :type format: a format name string, or Format object, or None to auto-detect
        :param id_tag: SD tag containing the record id
        :type id_tag: string, or None to use the record title
        :param reader_args: reader parameters passed to the underlying toolkit
        :type reader_args: a dictionary
        :param errors: specify how to handle errors
        :type errors: one of "strict", "report", or "ignore"
        :param location: object used to track parser state information
        :type location: a Location object, or None
        :returns: a :class:`chemfp.FingerprintIterator` which iterates over the (id, fingerprint) pair
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def read_molecule_fingerprints_from_string(self, content, format=None, id_tag=None,
                                                reader_args=None, errors="strict", location=None):
        """Read fingerprints from structure records in a string, as a FingerprintIterator

        Iterate through the *format* structure records in *content*. Use the
        fingerprint type to compute the fingerprint. For SD files, use *id_tag*
        to get the record id from the given SD tag instead of the title line.

        The *reader_args* dictionary parameters depend on the toolkit and format.
        For details see the docstring for ``self.toolkit.read_molecules``.

        The *errors* parameter specifies how to handle errors. "strict" raises
        an exception, "report" sends a message to stderr and goes to the next
        record, and "ignore" goes to the next record.

        The *location* parameter takes a Location instance. If None then a default
        Location will be created.
        
        :param content: the string containing structure records
        :type source: a string
        :param format: the input structure format
        :type format: a format name string, or Format object
        :param id_tag: SD tag containing the record id
        :type id_tag: string, or None to use the record title
        :param reader_args: reader parameters passed to the underlying toolkit
        :type reader_args: a dictionary
        :param errors: specify how to handle errors
        :type errors: one of "strict", "report", or "ignore"
        :param location: object used to track parser state information
        :type location: a Location object, or None
        :returns: a :class:`chemfp.FingerprintIterator` which iterates over the (id, fingerprint) pair
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def parse_molecule_fingerprint(self, content, format, reader_args=None, errors="strict"):
        """Parse the first molecule record of the content then compute and return the fingerprint

        Read the first molecule from *content*, which contains records
        in the given *format*. Compute and return its fingerprint.

        The *reader_args* dictionary parameters depend on the toolkit and format.
        For details see the docstring for ``self.toolkit.read_molecules``.

        The *errors* parameter specifies how to handle errors. "strict" raises
        an exception, "report" sends a message to stderr and return None for
        the fingerprint, and "ignore" returns None for the fingerprint without
        any extra message.
        
        :param content: the string containing at least one structure record
        :type source: a string
        :param format: the input structure format
        :type format: a format name string, or Format object
        :param reader_args: reader parameters passed to the underlying toolkit
        :type reader_args: a dictionary
        :param errors: specify how to handle errors
        :type errors: one of "strict", "report", or "ignore"
        :returns: the fingerprint as a byte string
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def parse_id_and_molecule_fingerprint(self, content, format, id_tag=None,
                                          reader_args=None, errors="strict"):
        """Parse the first molecule record of the content then compute and return the id and fingerprint

        Read the first molecule from *content*, which contains records in
        the given *format*. Compute its fingerprint and get the molecule id.
        For an SD record use *id_tag* to get the record id from the given
        SD tag instead of from the title line.

        Return the id and fingerprint as the (id, fingerprint) pair.
        
        The *reader_args* dictionary parameters depend on the toolkit and format.
        For details see the docstring for ``self.toolkit.read_molecules``.

        The *errors* parameter specifies how to handle errors. "strict" raises
        an exception, "report" sends a message to stderr and return None for
        values it cannot compute, and "ignore" is like "report" but without
        the error message. For "report" and "ignore", if the molecule cannot
        be parsed then the result will be (None, None). If the fingerprint
        cannot be computed then the result will be (id, None).

        :param content: the string containing at least one structure record
        :type source: a string
        :param format: the input structure format
        :type format: a format name string, or Format object
        :param id_tag: SD tag containing the record id
        :type id_tag: string, or None to use the record title
        :param reader_args: reader parameters passed to the underlying toolkit
        :type reader_args: a dictionary
        :param errors: specify how to handle errors
        :type errors: one of "strict", "report", or "ignore"
        :returns: a pair of (id string, fingerprint byte string)
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def make_id_and_molecule_fingerprint_parser(self, format, id_tag=None, reader_args=None, errors="strict"):
        """Make a function which parses molecule from a record and returns the id and computed fingerprint

        This is a very specialized function, designed for performance, but it
        doesn't appear to give any advantage. You likely don't need it.

        Return a function which parses a content string containing structure
        records in the given *format* to get a molecule. Use the molecule to
        compute the fingerprint and get its id. For an SD record use *id_tag*
        to get the record id from the given SD tag instead of from the
        title line.

        The new function will return the (id, fingerprint) pair.
        
        The *reader_args* dictionary parameters depend on the toolkit and format.
        For details see the docstring for ``self.toolkit.read_molecules``.

        The *errors* parameter specifies how to handle errors. "strict" raises
        an exception, "report" sends a message to stderr and return None for
        values it cannot compute, and "ignore" is like "report" but without
        the error message. For "report" and "ignore", if the molecule cannot
        be parsed then the result will be (None, None). If the fingerprint
        cannot be computed then the result will be (id, None).

        :param format: the input structure format
        :type format: a format name string, or Format object
        :param id_tag: SD tag containing the record id
        :type id_tag: string, or None to use the record title
        :param reader_args: reader parameters passed to the underlying toolkit
        :type reader_args: a dictionary
        :param errors: specify how to handle errors
        :type errors: one of "strict", "report", or "ignore"
        :returns: a function which takes a content string and returns an (id, fingerprint) pair
        """

    def compute_fingerprint(self, mol):
        """Compute and return the fingerprint byte string for the toolkit molecule

        :param mol: a toolkit molecule
        :returns: the fingerprint as a byte string
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def compute_fingerprints(self, mols):
        """Compute and return the fingerprint for each toolkit molecule in an iterator

        This function is a slightly optimized version of::

          for mol in mols:
            yield self.compute_fingerprint(mol)

        :param mols: an iterable of toolkit molecules
        :returns: a generator of fingerprints, one per molecule
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def get_fingerprint_family(self):
        """Return the fingerprint family for this fingerprint type

        :returns: a :class:`FingerprintFamily`
        """
        return registry.get_family(self.name)

def _compute_molecule_fingerprints(reader, fingerprinter):
    for id, mol in reader:
        if mol is None:
            yield id, None
        else:
            yield id, fingerprinter(mol)

class ThreadsafeFingerprinterMixin(object):
    def __init__(self, fingerprint_kwargs):
        super(ThreadsafeFingerprinterMixin, self).__init__(fingerprint_kwargs)
        self._fingerprinter = self._make_fingerprinter(**fingerprint_kwargs)

    def make_fingerprinter(self):
        return self._fingerprinter

    make_fingerprinter.__doc__ = BaseFingerprintType.__doc__

class FingerprintType(BaseFingerprintType):
    """The base to all fingerprint types

    A fingerprint type has the following public attributes:

    .. py:attribute:: name

      the fingerprint name, including the version

    .. py:attribute:: base_name

      the fingerprint name, without the version

    .. py:attribute:: version

      the fingerprint version

    .. py:attribute:: toolkit

       the toolkit API for the underlying chemistry toolkit, or None

    .. py:attribute:: software

       a string which characterizes the toolkit, including version information

    .. py:attribute:: num_bits

       the number of bits in this fingerprint type
       
    .. py:attribute:: fingerprint_kwargs

       a dictionary of the fingerprint arguments

    The built-in fingerprint types are:

    * :class:`chemfp.openbabel_types.OpenBabelFP2FingerprintType_v1` - ``OpenBabel-FP2/1`` -
      Open Babel FP2
    * :class:`chemfp.openbabel_types.OpenBabelFP3FingerprintType_v1` - ``OpenBabel-FP3/1`` -
      Open Babel FP3
    * :class:`chemfp.openbabel_types.OpenBabelFP4FingerprintType_v1` - ``OpenBabel-FP4/1`` -
      Open Babel FP4
    * :class:`chemfp.openbabel_types.OpenBabelMACCSFingerprintType_v1` - ``OpenBabel-MACCS/1`` -
      Open Babel 166 MACCS keys
    * :class:`chemfp.openbabel_types.OpenBabelMACCSFingerprintType_v2` - ``OpenBabel-MACCS/2`` -
      Open Babel 166 MACCS keys
    * :class:`chemfp.openbabel_patterns.SubstructOpenBabelFingerprinter_v1` - ``ChemFP-Substruct-OpenBabel/1`` -
      chemfp's 881 CACTVS/PubChem-like keys implemented with Open Babel
    * :class:`chemfp.openbabel_patterns.RDMACCSOpenBabelFingerprinter_v1` - ``RDMACCS-OpenBabel/1`` -
      chemfp's own 166 MACCS keys implemented with Open Babel (does not include key 44)
    * :class:`chemfp.openbabel_patterns.RDMACCSOpenBabelFingerprinter_v2` - ``RDMACCS-OpenBabel/1`` -
      chemfp's own 166 MACCS keys implemented with Open Babel

    * :class:`chemfp.openeye_types.OpenEyeCircularFingerprintType_v2` - ``OpenEye-Circular/2`` -
      OEGraphSim circular fingerprints
    * :class:`chemfp.openeye_types.OpenEyeMACCSFingerprintType_v2` - ``OpenEye-MACCS166/2`` -
      OEGraphSim 166 MACCS keys
    * :class:`chemfp.openeye_types.OpenEyePathFingerprintType_v2` - ``OpenEye-Path/2`` -
      OEGraphSim path fingerprints
    * :class:`chemfp.openeye_types.OpenEyeTreeFingerprintType_v2` - ``OpenEye-Tree/2`` -
      OEGraphSim tree fingerprints

    * :class:`chemfp.openeye_patterns.SubstructOpenEyeFingerprinter_v1` - ``ChemFP-Substruct-OpenEye/1`` -
      chemfp's 881 CACTVS/PubChem-like keys implemented with OEChem
    * :class:`chemfp.openeye_patterns.RDMACCSOpenEyeFingerprinter_v1` - ``RDMACCS-OpenEye/1`` -
      chemfp's own 166 MACCS keys implemented with OEChem (does not include key 44)
    * :class:`chemfp.openeye_patterns.RDMACCSOpenEyeFingerprinter_v2` - ``RDMACCS-OpenEye/2`` -
      chemfp's own 166 MACCS keys implemented with OEChem

    * :class:`chemfp.rdkit_types.RDKitFingerprintType_v1` - RDKit-Fingerprint/1 - RDKit path and tree fingerprint
    * :class:`chemfp.rdkit_types.RDKitFingerprintType_v2` - RDKit-Fingerprint/2 - RDKit path and tree fingerprint
    * :class:`chemfp.rdkit_types.RDKitMACCSFingerprintType_v1` - ``RDKit-MACCS/1`` -
      RDKit 166 MACCS keys (does not include key 44)
    * :class:`chemfp.rdkit_types.RDKitMACCSFingerprintType_v2` - ``RDKit-MACCS/2`` -
      RDKit 166 MACCS keys
    * :class:`chemfp.rdkit_types.RDKitMorganFingerprintType_v1` - ``RDKit-Morgan/1`` -
      RDKit circular fingerprints
    * :class:`chemfp.rdkit_types.RDKitAtomPairFingerprint_v1` - ``RDKit-AtomPair/1`` -
      RDKit atom pair fingerprints
    * :class:`chemfp.rdkit_types.RDKitAtomPairFingerprint_v2` - ``RDKit-AtomPair/2`` -
      RDKit atom pair fingerprints
    * :class:`chemfp.rdkit_types.RDKitTorsionFingerprintType_v1` - ``RDKit-Torsion/1`` -
      RDKit torsion fingerprints
    * :class:`chemfp.rdkit_types.RDKitTorsionFingerprintType_v2` - ``RDKit-Torsion/2`` -
      RDKit torsion fingerprints
    * :class:`chemfp.rdkit_types.RDKitTorsionFingerprintType_v3` - ``RDKit-Torsion/3`` -
      RDKit torsion fingerprints

    * :class:`chemfp.rdkit_patterns.SubstructRDKitFingerprintType_v1` - ``ChemFP-Substruct-RDKit/1`` -
      chemfp's 881 CACTVS/PubChem-like keys implemented with RDKit
    * :class:`chemfp.rdkit_patterns.RDMACCSRDKitFingerprinter_v1` - ``RDMACCS-RDKit/1`` -
      chemfp's own 166 MACCS keys implemented with OEChem (does not include key 44)
    * :class:`chemfp.rdkit_patterns.RDMACCSRDKitFingerprinter_v2` - ``RDMACCS-RDKit/2`` -
      chemfp's own 166 MACCS keys implemented with OEChem


    """
    toolkit = None
    software = None
    _arguments = None
    _format_string = None
    _make_fingerprinter = None

    

    # Fingerprint methods
    def get_type(self):
        if self.fingerprint_kwargs:
            encoded_args = self._arguments.encode_fingerprint_args(self.fingerprint_kwargs)
            return self.name + " " + self._format_string.format(**encoded_args)
        else:
            return self.name

    def make_fingerprinter(self):
        # I assume the fingerprinter contains state, so always make a new one.
        # (Use the Threadsafe mixin if it's thread-safe.)
        return self._make_fingerprinter(**self.fingerprint_kwargs)


    def get_metadata(self, sources=None):
        """Return a Metadata appropriate for the given fingerprint type.

        This is most commonly used to make a :class:`chemfp.Metadata`
        that can be passed into a :class:`chemfp.FingerprintWriter`.
        
        If *sources* is a string or a list of strings then it will passed
        to the newly created Metadata instance. It should contain filenames
        or other description of the fingerprint sources.

        :param sources: fingerprint source filenames or other description
        :type sources: None, a string, or list of strings
        :returns: a :class:`chemfp.Metadata`
        """
        from chemfp import Metadata, io
        return Metadata(num_bits = self.num_bits,
                        sources = sources,
                        software = self.software,
                        type = self.get_type(),
                        date = io.utcnow())

        
    def read_molecule_fingerprints(self, source, format=None, id_tag=None,
                                    reader_args=None, errors="strict", location=None):
        """Read fingerprints from a structure source. Returns a stream of (id, fp) pairs"""
        from chemfp import FingerprintIterator
        reader = self.toolkit.read_ids_and_molecules(source, format, id_tag, reader_args, errors, location)
        fingerprinter = self.make_fingerprinter()
        return FingerprintIterator(self.get_metadata(source),
                                   _compute_molecule_fingerprints(reader, fingerprinter),
                                   reader.location, reader.close)

    def read_molecule_fingerprints_from_string(self, content, format=None, id_tag=None,
                                                reader_args=None, errors="strict", location=None):
        """Read fingerprints from a string. Returns a stream of (id, fp) pairs"""
        from chemfp import FingerprintIterator
        reader = self.toolkit.read_ids_and_molecules_from_string(content, format, id_tag, reader_args,
                                                                 errors, location)
        fingerprinter = self.make_fingerprinter()
        return FingerprintIterator(self.get_metadata(),
                                   _compute_molecule_fingerprints(reader, fingerprinter),
                                   reader.location, reader.close)
                                

    def parse_molecule_fingerprint(self, content, format, reader_args=None, errors="strict"):
        """Return the corresponding fingerprint byte string for the structure record"""
        mol = self.toolkit.parse_molecule(content, format, reader_args=reader_args, errors=errors)
        if mol is None:
            return None
        fingerprinter = self.make_fingerprinter()
        return fingerprinter(mol)

    def parse_id_and_molecule_fingerprint(self, content, format, id_tag=None,
                                          reader_args=None, errors="strict"):
        """Return the corresponding id and fingerprint byte string for the structure record"""
        id, mol = self.toolkit.parse_id_and_molecule(content, format, id_tag=id_tag,
                                                     reader_args=reader_args, errors=errors)
        if mol is None:
            return id, None
        fingerprinter = self.make_fingerprinter()
        return id, fingerprinter(mol)

    def make_id_and_molecule_fingerprint_parser(self, format, id_tag=None, reader_args=None, errors="strict"):
        """Make a function which parses molecule from a record and returns the id and computed fingerprint"""
        parse_id_and_molecule = self.toolkit.make_id_and_molecule_parser(
            format=format, id_tag=id_tag, reader_args=reader_args, errors=errors)
        fingerprinter = self.make_fingerprinter()

        def parse_id_and_molecule_fingerprint(content):
            id, mol = parse_id_and_molecule(content)
            if mol is not None:
                return id, fingerprinter(mol)
            else:
                return id, None

        return parse_id_and_molecule_fingerprint

    def compute_fingerprint(self, mol):
        """Return the corresponding fingerprint byte string for the molecule"""
        fingerprinter = self.make_fingerprinter()
        return fingerprinter(mol)

    def compute_fingerprints(self, mols):
        """Return the corresponding fingerprint byte strings for a stream of molecules"""
        fingerprinter = self.make_fingerprinter()
        for mol in mols:
            yield fingerprinter(mol)

    # XXX Hack because I don't want to spend the time to remove the BaseFingerprint
    
    get_type.__doc__ = BaseFingerprintType.get_type.__doc__
    make_fingerprinter.__doc__ = BaseFingerprintType.make_fingerprinter.__doc__
    read_molecule_fingerprints.__doc__ = BaseFingerprintType.read_molecule_fingerprints.__doc__
    read_molecule_fingerprints_from_string.__doc__ = BaseFingerprintType.read_molecule_fingerprints_from_string.__doc__
    parse_molecule_fingerprint.__doc__ = BaseFingerprintType.parse_molecule_fingerprint.__doc__
    parse_id_and_molecule_fingerprint.__doc__ = BaseFingerprintType.parse_id_and_molecule_fingerprint.__doc__
    make_id_and_molecule_fingerprint_parser.__doc__ = BaseFingerprintType.make_id_and_molecule_fingerprint_parser.__doc__
    compute_fingerprint.__doc__ = BaseFingerprintType.compute_fingerprint.__doc__
    compute_fingerprints.__doc__ = BaseFingerprintType.compute_fingerprints.__doc__

########
def _get_module(module_path):
    if isinstance(module_path, _compat.string_types):
        module = __import__(module_path)
        for name in module_path.split(".")[1:]:
            module = getattr(module, name)
    else:
        module = module_path
    return module


# Part of the code uses a a list of (key, value) tuples. Other parts
# use a dict API and expect to call .items() to get those
# tuples. FakeSortedDict is an adapter of the first to the second.

class FakeSortedDict(object):
    def __init__(self, items):
        self._items = items
    def __nonzero__(self):  # Python 2.x
        return bool(self._items)
    def __bool__(self):  # Python 3.x
        return bool(self._items)
    def items(self):
        return self._items
    def iteritems(self):
        return self._items
    def __repr__(self):
        return "FakeSortedDict(%r)" % (self._items,)


# Only the "register_*" methods are part of the public API.

class FamilyRegistry(object):
    def __init__(self):
        self._factories = {}  # The available factories
        self._unavailable_reasons = {}  # Reason for why a fingerprint isn't available
        self._loaded_modules = set()
        self._lazy_names = {}

    @property
    def factories(self):
        self._resolve_all()
        return self._factories

    def _get_factories(self, toolkit_name=None):
        if toolkit_name is None:
            return self.factories
        else:
            self._resolve_all(toolkit_name)
            return dict( (k,v) for (k, v) in self._factories.items()
                                if v.toolkit is not None and v.toolkit.name == toolkit_name)

    def register_module(self, module_path, names, required_toolkit):
        if not isinstance(module_path, _compat.string_types):
            raise TypeError("module_path must be a module path string")
        if isinstance(names, _compat.string_types):
            names = [names]
        for name in names:
            self._lazy_names[name] = (module_path, required_toolkit)

    def _resolve_name(self, name):
        input_name = name
        if name not in self._lazy_names:
            name = name.partition("/")[0]
            if name not in self._lazy_names:
                return
        
        module_path, required_toolkit = self._lazy_names[name]
        if module_path in self._loaded_modules:
            return

        if required_toolkit is not None:
            if not _has_toolkit(required_toolkit):
                self.register_unavailable(input_name, "toolkit %r is not available" % (required_toolkit,))
                return

        try:
            module = _get_module(module_path)
        except ImportError as err:
            raise ValueError("Cannot resolve fingerprint %r because importing %r failed: %s"
                             % (input_name, module_path, err))
        module.resolve_chemfp_fingerprint_types(self)
        self._loaded_modules.add(module_path)
        del self._lazy_names[name]

        if (name not in self._factories and
            name not in self._unavailable_reasons):
            raise AssertionError("Fingerprint %r should have been registered by module %r but wasn't"
                                 % (name, module_path,))

    def _resolve_all(self, toolkit_name=None):
        can_skip = set()
        for loop in range(10):
            if set(self._lazy_names) == can_skip:
                return

            for name, (module_path, required_toolkit) in list(self._lazy_names.items()):
                if toolkit_name is not None and required_toolkit != toolkit_name:
                    can_skip.add(name)
                    continue
                if module_path not in self._loaded_modules:
                    if required_toolkit is not None and not _has_toolkit(required_toolkit):
                        self.register_unavailable(name, "toolkit %r is not available" % (required_toolkit,))
                    else:
                        try:
                            module = _get_module(module_path)
                        except ImportError as err:
                            self.register_unavailable(
                                name,
                                "Cannot resolve fingerprint %r because importing %r failed: %s"
                                % (name, module_path, err))
                        else:
                            module.resolve_chemfp_fingerprint_types(self)
                            self._loaded_modules.add(module_path)

                if (name not in self._factories and
                    name not in self._unavailable_reasons):
                    raise AssertionError(
                        "Fingerprint %r should have been registered by module %r but wasn't"
                        % (name, module_path,))
                
                del self._lazy_names[name]

        raise AssertionError("Could not resolve types %r. Is there a registration cycle?"
                             % (sorted(self._lazy_names),))
    
    def register_family(self, fingerprint_family):
        name = fingerprint_family.name
        if name is None:
            raise ValueError("The fingerprint family must specify a name")
        base_name, version = _verify_name(name)

        # Always change based on the given name
        self._factories[name] = fingerprint_family
        if name in self._unavailable_reasons:
            del self._unavailable_reasons[name]

        # Don't change the base name if one already exists
        if base_name not in self._factories:
            self._factories[base_name] = fingerprint_family

            # Remove any "unavailable" reason, since it can be resolved
            if base_name in self._unavailable_reasons:
                del self._unavailable_reasons[base_name]

        # Used during debugging to check that I registered everything
        if 0 and base_name not in self._lazy_names:
            import sys
            sys.stderr.write("Unregistered base_name: %r\n" % (base_name,))

    def register_fingerprint_type(self, fingerprint_type):
        family = FingerprintFamily(fingerprint_type)
        self.register_family(family)

    def register_unavailable(self, name, reason):
        base_name, version = _verify_name(name)

        # Always change based on the given name
        self._unavailable_reasons[name] = reason
        if name in self._factories:
            del self._factories[name]

        # Don't replace the base name if it already exists
        if base_name not in self._factories:
            self._unavailable_reasons[base_name] = reason

    def get_all_fingerprint_names(self):
        self._resolve_all()
        x = set(self.factories)
        x.update(self._unavailable_reasons)
        return x
        

    def get_family(self, typename):
        self._resolve_name(typename)
        try:
            return self._factories[typename]
        except KeyError:
            reason = self._unavailable_reasons.get(typename, None)
            if reason is None:
                err = FingerprintTypeValueError("Unknown fingerprint type %s" % (_compat.myrepr(typename),))
            else:
                err = FingerprintTypeValueError("Unable to use %s: %s" % (typename, reason))
        raise err

    def get_families(self, toolkit_name=None):
        # Return only versioned families
        factories = self._get_factories(toolkit_name)
        return [v for (k, v) in sorted(factories.items()) if "/" not in k]

    def get_family_names(self, include_unavailable, toolkit_name=None):
        factories = self._get_factories(toolkit_name)
        names = set(factories)
        if include_unavailable:
            if toolkit_name is None:
                names.update(self._unavailable_reasons)
            else:
                raise NotImplementedError("cannot yet specify both include_unavailable and toolkit_name")
        return set(name for name in names if "/" not in name)  # only return versioned names
 
    def has_family(self, typename):
        if typename in self._factories:
            return True
        self._resolve_name(typename)
        return typename in self._factories
    
    def get_fingerprint_type(self, type, fingerprint_kwargs=None):
        parsed = tokenize_type_string(type)
        name = parsed.name
        self._resolve_name(name)
        err = None
        try:
            family = self._factories[name]
        except KeyError:
            reason = self._unavailable_reasons.get(name, None)
            if reason is None:
                err = FingerprintTypeValueError("Unknown fingerprint type %s" % (_compat.myrepr(name),))
            else:
                err = FingerprintTypeValueError("Unable to use %s: %s" % (name, reason))
        if err is not None:
            # Don't chain exceptions
            raise err

        value_error = None
        try:
            final_kwargs = family.get_kwargs_from_text_settings(FakeSortedDict(parsed.settings))
        except ValueError as err:
            # This is an odd way to avoid exception chaining. I should say "from None" but
            # that doesn't work under Python 2.7.
            value_error = FingerprintTypeValueError("Cannot get fingerprint type %r: %s" % (type, err))
        if value_error is not None:
            raise value_error

        if fingerprint_kwargs:
            final_kwargs.update(fingerprint_kwargs)
        try:
            return family.from_kwargs(final_kwargs)
        except ValueError as err:
            msg = "Cannot get fingerprint type %r: %s" % (type, err)
        raise FingerprintTypeValueError(msg)
            

    def get_fingerprint_type_from_text_settings(self, type, settings):
        parsed = tokenize_type_string(type)
        name = parsed.name
        self._resolve_name(name)
        err = None
        try:
            family = self._factories[name]
        except KeyError:
            reason = self._unavailable_reasons.get(name, None)
            if reason is None:
                err = FingerprintTypeValueError("Unknown fingerprint type %s" % (_compat.myrepr(name),))
            else:
                err = FingerprintTypeValueError("Unable to use %s: %s" % (name, reason))
        if err is not None:
            # Don't chain exceptions
            raise err

        try:
            type_kwargs = family.get_kwargs_from_text_settings(FakeSortedDict(parsed.settings))
        except ValueError as err:
            raise FingerprintTypeValueError("Error with type %r: %s" % (type, err))

        if settings:
            # This is a bit complicated. The "get_kwargs_from_text_settings" returns a complete
            # dictionary, with defaults filled in, when I only want the new terms.
            try:
                settings_kwargs = family.get_kwargs_from_text_settings(settings)
                value_err = None
            except ValueError as err:
                # Do a bit of oddness to support Python 2 and Python 3 and
                # suppress exception chaining on Python 3.
                value_err = FingerprintTypeValueError("Error with type %r: %s" % (type, err))
            if value_err is not None:
                raise value_err

            # Override the type string values using only the new values in settings
            # and not the defaults.
            for name in settings:
                type_kwargs[name] = settings_kwargs[name]

        return family(**type_kwargs)



def resolve_chemfp_fingerprint_types(registry):
    register_module = registry.register_module
    
    register_module("chemfp.rdkit_types", "RDKit-MACCS166", "rdkit")
    register_module("chemfp.rdkit_types", "RDKit-Fingerprint", "rdkit")
    register_module("chemfp.rdkit_types", "RDKit-Morgan", "rdkit")
    register_module("chemfp.rdkit_types", "RDKit-Torsion", "rdkit")
    register_module("chemfp.rdkit_types", "RDKit-AtomPair", "rdkit")
    register_module("chemfp.rdkit_types", "RDKit-Pattern", "rdkit")
    register_module("chemfp.rdkit_types", "RDKit-Avalon", "rdkit")
    register_module("chemfp.rdkit_types", "RDKit-SECFP", "rdkit")
    register_module("chemfp.rdkit_patterns", "RDMACCS-RDKit", "rdkit")
    register_module("chemfp.rdkit_patterns", "ChemFP-Substruct-RDKit", "rdkit")

    register_module("chemfp.openeye_types", "OpenEye-MACCS166", "openeye")
    register_module("chemfp.openeye_types", "OpenEye-Path", "openeye")
    register_module("chemfp.openeye_types", "OpenEye-Circular", "openeye")
    register_module("chemfp.openeye_types", "OpenEye-Tree", "openeye")
    register_module("chemfp.openeye_types", "OpenEye-MoleculeScreen", "openeye")
    register_module("chemfp.openeye_types", "OpenEye-SMARTSScreen", "openeye")
    register_module("chemfp.openeye_types", "OpenEye-MDLScreen", "openeye")
    register_module("chemfp.openeye_patterns", "RDMACCS-OpenEye", "openeye")
    register_module("chemfp.openeye_patterns", "ChemFP-Substruct-OpenEye", "openeye")

    register_module("chemfp.openbabel_types", "OpenBabel-FP2", "openbabel")
    register_module("chemfp.openbabel_types", "OpenBabel-FP3", "openbabel")
    register_module("chemfp.openbabel_types", "OpenBabel-FP4", "openbabel")
    register_module("chemfp.openbabel_types", "OpenBabel-MACCS", "openbabel")
    register_module("chemfp.openbabel_types", "OpenBabel-ECFP0", "openbabel")
    register_module("chemfp.openbabel_types", "OpenBabel-ECFP2", "openbabel")
    register_module("chemfp.openbabel_types", "OpenBabel-ECFP4", "openbabel")
    register_module("chemfp.openbabel_types", "OpenBabel-ECFP6", "openbabel")
    register_module("chemfp.openbabel_types", "OpenBabel-ECFP8", "openbabel")
    register_module("chemfp.openbabel_types", "OpenBabel-ECFP10", "openbabel")
    register_module("chemfp.openbabel_patterns", "RDMACCS-OpenBabel", "openbabel")
    register_module("chemfp.openbabel_patterns", "ChemFP-Substruct-OpenBabel", "openbabel")


registry = FamilyRegistry()
resolve_chemfp_fingerprint_types(registry)

def parse_type(name):
    import warnings
    warnings.warn("Please use chemfp.get_fingerprint_type() instead of chemfp.types.parse_type()")
    return registry.get_fingerprint_type(name)
